# TID 251 enforces to not import from nested irdl
# We need to skip it here to allow importing from here instead.
from .attributes import *  # noqa: TID251
from .constraints import *  # noqa: TID251
from .error import *  # noqa: TID251
from .operations import *  # noqa: TID251
